"""
Module for implementing helper functions.
"""
from numbers import Number
try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping
from collections import defaultdict
import numpy as np
import sympy as sp
from scipy.fftpack import dct
from shenfun.optimization import runtimeoptimizer
from shenfun.config import config
from .findbasis import get_bc_basis, get_stencil_matrix, n

__all__ = ['dx', 'clenshaw_curtis1D', 'CachedArrayDict', 'surf3D',
           'wrap_periodic', 'outer', 'dot', 'apply_mask', 'integrate_sympy',
           'mayavi_show', 'quiver3D', 'get_bc_basis', 'get_stencil_matrix',
           'scalar_product', 'n', 'cross', 'reset_profile']

def dx(u, weighted=False):
    r"""Compute integral of u over domain

    .. math::

        \int_{\Omega} u dx

    Parameters
    ----------

    u : Array
        The Array to integrate

    Note
    ----
    This function assumes a standard reference domain. If the domain
    is not standard, then modify the result accordingly.

    """
    T = u.function_space()
    dim = len(u.shape)
    if dim == 1:
        w = T.points_and_weights(weighted=weighted)[1]
        return np.sum(u*w).item()

    uc = u.copy()
    for ax in range(dim):
        uc = uc.redistribute(axis=ax)
        w = T.bases[ax].points_and_weights(weighted=weighted)[1]
        sl = [np.newaxis]*len(uc.shape)
        sl[ax] = slice(None)
        uu = np.sum(uc*w[tuple(sl)], axis=ax)
        sl = [slice(None)]*len(uc.shape)
        sl[ax] = np.newaxis
        uc[:] = uu[tuple(sl)]
    return uc.flat[0]

def clenshaw_curtis1D(u, quad="GC"):  # pragma: no cover
    """Clenshaw-Curtis integration in 1D"""
    assert u.ndim == 1
    N = u.shape[0]
    if quad == 'GL':
        w = np.arange(0, N, 1, dtype=float)
        w[2:] = 2./(1-w[2:]**2)
        w[0] = 1
        w[1::2] = 0
        ak = dct(u, 1)
        ak /= (N-1)
        return np.sqrt(np.sum(ak*w))

    assert quad == 'GC'
    d = np.zeros(N)
    k = 2*(1 + np.arange((N-1)//2))
    d[::2] = (2./N)/np.hstack((1., 1.-k*k))
    w = dct(d, type=3)
    return np.sqrt(np.sum(u*w))

class CachedArrayDict(MutableMapping):
    """Dictionary for caching Numpy arrays (work arrays)

    Example
    -------

    >>> import numpy as np
    >>> from shenfun.utilities import CachedArrayDict
    >>> work = CachedArrayDict()
    >>> a = np.ones((3, 4), dtype=int)
    >>> w = work[(a, 0, True)] # create work array with shape as a
    >>> print(w.shape)
    (3, 4)
    >>> print(w)
    [[0 0 0 0]
     [0 0 0 0]
     [0 0 0 0]]
    >>> w2 = work[(a, 1, True)] # Get different(note 1!) array of same shape/dtype
    """
    def __init__(self):
        self._data = {}

    def __getitem__(self, key):
        newkey, fill = self.__keytransform__(key)
        try:
            value = self._data[newkey]
        except KeyError:
            shape, dtype, _ = newkey
            value = np.empty(shape, dtype=np.dtype(dtype, align=True))
            self._data[newkey] = value
        if fill:
            value.fill(0)
        return value

    @staticmethod
    def __keytransform__(key):
        assert len(key) == 3
        return (key[0].shape, key[0].dtype, key[1]), key[2]

    def __len__(self):
        return len(self._data)

    def __setitem__(self, key, value):
        self._data[self.__keytransform__(key)[0]] = value

    def __delitem__(self, key):
        del self._data[self.__keytransform__(key)[0]]

    def __iter__(self):
        return iter(self._data)

    def values(self):
        raise TypeError('Cached work arrays not iterable')

def reset_profile(prof):
    """Reset profiler for kernprof

    Parameters
    ----------
    prof : The profiler
    """
    prof.code_map = {}
    prof.last_time = {}
    prof.enable_count = 0
    for func in prof.functions:
        prof.add_function(func)

def cross(c, a, b):
    """Cross product c = a x b

    Parameters
    ----------
    c : Array
    a : Array
    b : Array

    Returns
    -------
    c : Array
    """
    if a.ndim == 3:
        cross2D(c, a, b)
    elif a.ndim == 4:
        cross3D(c, a, b)
    else:
        crossND(c, a, b)
    return c

@runtimeoptimizer
def cross2D(c, a, b):
    c[:] = a[0]*b[1]-a[1]*b[0]

@runtimeoptimizer
def cross3D(c, a, b):
    c[0] = a[1]*b[2] - a[2]*b[1]
    c[1] = a[2]*b[0] - a[0]*b[2]
    c[2] = a[0]*b[1] - a[1]*b[0]

@runtimeoptimizer
def crossND(c, a, b):
    cross3D(c, a, b)

def outer(a, b, c):
    r"""Return outer product $c_{i,j} = a_i b_j$

    Parameters
    ----------
    a : Array of shape (N, ...)
    b : Array of shape (N, ...)
    c : Array of shape (N*N, ...)

    The outer product is taken over the first index of a and b,
    for all remaining indices.
    """
    av = a.v
    bv = b.v
    cv = c.v
    symmetric = a is b
    if av.shape[0] == 2:
        outer2D(av, bv, cv, symmetric)
    elif av.shape[0] == 3:
        outer3D(av, bv, cv, symmetric)
    return c

@runtimeoptimizer
def outer2D(a, b, c, symmetric):
    c[0] = a[0]*b[0]
    c[1] = a[0]*b[1]
    if symmetric:
        c[2] = c[1]
    else:
        c[2] = a[1]*b[0]
    c[3] = a[1]*b[1]

@runtimeoptimizer
def outer3D(a, b, c, symmetric):
    c[0] = a[0]*b[0]
    c[1] = a[0]*b[1]
    c[2] = a[0]*b[2]
    if symmetric:
        c[3] = c[1]
        c[6] = c[2]
        c[7] = c[5]
    else:
        c[3] = a[1]*b[0]
        c[6] = a[2]*b[0]
        c[7] = a[2]*b[1]
    c[4] = a[1]*b[1]
    c[5] = a[1]*b[2]
    c[8] = a[2]*b[2]

def dot(u, v, output_array=None, forward_output=True):
    """Return dot product of u and v

    Parameters
    ----------
    u : Function or Array
    v : Function or Array
    output_array : Function or Array
        Must be of correct shape
    forward_output : bool, optional
        Return the result as a Function if True, and an Array if
        False.

    Note
    ----
    This function uses the 3/2-rule for dealiasing all spaces if
    the input arrays are spectral Functions. The returned Function
    is projected to an orthogonal space, without boundary conditions.

    """
    from shenfun import Function, Array
    assert isinstance(u, (Function, Array))
    assert isinstance(v, (Function, Array))
    Vu = u.function_space()
    Vv = v.function_space()
    assert Vu.dimensions == Vv.dimensions
    D = Vu.dimensions
    Vup = Vu
    Vvp = Vv

    # Get the correct output_array
    # Logic is to first follow the output_array if given,
    # use input_arrays if not
    uv = output_array
    if isinstance(output_array, Function):
        forward_output = True
        Top = output_array.function_space()
        uv = Array(Top)

    elif isinstance(output_array, Array):
        forward_output = False
        Top = output_array.function_space()
        uv = output_array

    else:
        output_rank = Vu.tensor_rank+Vv.tensor_rank-2
        if output_rank == 0:
            To = Vu[0].get_orthogonal()
        elif output_rank == 1:
            if Vu.tensor_rank == 1:
                To = Vu.get_orthogonal()
            else:
                To = Vu[0].get_orthogonal()
        elif output_rank == 2:
            To = Vu.get_orthogonal()

        Top = To.get_dealiased()
        uv = Array(Top)
    uv.fill(0)

    if Top.is_padded:
        Vup = Vu.get_dealiased()
        Vvp = Vv.get_dealiased()

    if isinstance(u, Array):
        ua = u

    else:
        ua = Array(Vup)
        ua = Vup.backward(u, ua)

    if isinstance(v, Array):
        va = v
    else:
        va = Array(Vvp)
        va = Vvp.backward(v, va)

    if Vu.coors.is_cartesian:

        if Vu.tensor_rank == 1 and Vv.tensor_rank == 1:
            uv = np.sum(ua.v*va.v, axis=0, out=uv)

        elif Vu.tensor_rank == 2 and Vv.tensor_rank == 1:
            for i in range(D):
                ui = ua[i]
                for j in range(D):
                    uv[i] += ui[j]*va[j]

        elif Vu.tensor_rank == 1 and Vv.tensor_rank == 2:
            for i in range(D):
                ui = ua[i]
                vi = va[i]
                for j in range(D):
                    uv[j] += ui*vi[j]

        elif Vu.tensor_rank == 2 and Vv.tensor_rank == 2:
            for i in range(D):
                ui = ua[i]
                wi = uv[i]
                for j in range(D):
                    vj = va[j]
                    for k in range(D):
                        wi[k] += ui[j]*vj[k]

        else:
            raise NotImplementedError

    else:
        gij = Vu.coors.get_metric_tensor(config['basisvectors'])
        mesh = Vup.local_mesh(True)
        def measure(g):
            sym0 = g.free_symbols
            m = []
            for sym in sym0:
                j = 'xyzrs'.index(str(sym))
                m.append(mesh[j])
            return sp.lambdify(tuple(sym0), g)(*m)

        if Vu.tensor_rank == 1 and Vv.tensor_rank == 1:
            if Vu.coors.is_orthogonal:
                for i in range(D):
                    uv += ua[i]*va[i]*measure(gij[i, i])
            else:
                for i in range(D):
                    ui = ua[i]
                    for j in range(D):
                        g = gij[i, j]
                        if not g == 0:
                            uv += ui*va[j]*measure(g)

        elif Vu.tensor_rank == 2 and Vv.tensor_rank == 1:
            for i in range(D):
                ui = ua[i]
                for j in range(D):
                    for k in range(D):
                        g = gij[j, k]
                        if not g == 0:
                            uv[i] += ui[j]*va[k]*measure(g)

        elif Vu.tensor_rank == 1 and Vv.tensor_rank == 2:
            for i in range(D):
                ui = ua[i]
                for j in range(D):
                    vj = va[j]
                    g = gij[i, j]
                    if not g == 0:
                        ms = measure(g)
                        for k in range(D):
                            uv[k] += ui*vj[k]*ms

        elif Vu.tensor_rank == 2 and Vv.tensor_rank == 2:
            for i in range(D):
                ui = ua[i]
                wi = uv[i]
                for j in range(D):
                    for k in range(D):
                        g = gij[j, k]
                        if not g == 0:
                            vk = va[k]
                            ms = measure(g)
                            for l in range(D):
                                wi[l] += ui[j]*vk[l]*ms

        else:
            raise NotImplementedError

    if forward_output is True:

        if isinstance(output_array, Function):
            output_array = uv.forward(output_array)
            return output_array

        uv_hat = Function(To)
        uv_hat = uv.forward(uv_hat)
        return uv_hat

    return uv


@runtimeoptimizer
def apply_mask(u_hat, mask):
    if mask is not None:
        u_hat *= mask
    return u_hat

def integrate_sympy(f, d):
    """Exact definite integral using sympy

    Try to convert expression `f` to a polynomial before integrating.

    See sympy issue https://github.com/sympy/sympy/pull/18613 to why this is
    needed. Poly().integrate() is much faster than sympy.integrate() when applicable.

    Parameters
    ----------
    f : sympy expression
    d : 3-tuple
        First item the symbol, next two the lower and upper integration limits
    """
    try:
        p = sp.Poly(f, d[0]).integrate()
        return p(d[2]) - p(d[1])
    except sp.PolynomialError:
        #return sp.Integral(f, d).evalf()
        return sp.integrate(f, d).evalf()

def rsplit(measures):
    d = []
    def _split(ms):
        ms = sp.sympify(ms)
        if isinstance(ms, Number):
            d = {'coeff': ms}
        elif len(ms.free_symbols) == 0:
            d = {'coeff': ms.evalf()}
        else:
            d = sp.separatevars(ms.factor(), dict=True)
        if d is None:
            return d
        d = defaultdict(lambda: 1, {str(k): v for k, v in d.items()})
        dc = d['coeff']
        d['coeff'] = int(dc) if isinstance(dc, (sp.Integer, int)) else float(dc)
        return d

    ms = sp.sympify(measures)
    di = _split(ms)
    if di is not None:
        d += [di]
    else:
        ms = ms.expand()
        for arg in ms.args:
            d += split(arg)
    return d

def split(measures, expand=False):

    def _split(ms):
        ms = sp.sympify(ms)
        if isinstance(ms, Number):
            d = {'coeff': ms}
        elif len(ms.free_symbols) == 0:
            d = {'coeff': ms.evalf()}
        else:
            d = sp.separatevars(ms.factor(), dict=True)
        if d is None:
            raise RuntimeError('Could not split', ms)
        d = defaultdict(lambda: 1, {str(k): v for k, v in d.items()})
        dc = sp.sympify(d['coeff'])
        if dc.is_integer:
            d['coeff'] = int(dc)
        elif dc.is_real:
            d['coeff'] = float(dc)
        elif dc.is_complex:
            d['coeff'] = complex(dc)
        else:
            raise RuntimeError
        return d

    ms = sp.sympify(measures)
    if not expand:
        try:
            di = _split(ms)
            if di is not None:
                return [di]
        except RuntimeError:
            pass

    ms = ms.expand()
    result = []
    if isinstance(ms, sp.Add) or (isinstance(ms, sp.Mul) and np.all([isinstance(s, sp.Add) for s in ms.args])):
        for arg in ms.args:
            result.append(_split(arg))
    else:
        result.append(_split(ms))
    return result

def oldsplit(measures):

    def _split(mss, result):
        for ms in mss:
            ms = sp.sympify(ms)
            if isinstance(ms, sp.Mul):
                # Multiplication of two or more terms
                result = _split(ms.args, result)
                continue

            # Something else with only one symbol
            sym = ms.free_symbols
            assert len(sym) <= 1
            if len(sym) == 1:
                sym = sym.pop()
                result[str(sym)] *= ms
            else:
                ms = int(ms) if isinstance(ms, sp.Integer) else float(ms)
                result['coeff'] *= ms
        return result

    ms = sp.sympify(measures).expand()
    result = []
    if isinstance(ms, sp.Add):
        for arg in ms.args:
            result.append(_split([arg], defaultdict(lambda: 1)))
    else:
        result.append(_split([ms], defaultdict(lambda: 1)))
    return result

def mayavi_show():
    """
    Return show function that updates the mayavi figure in the background.
    """
    from pyface.api import GUI
    from mayavi import mlab
    return mlab.show(GUI().stop_event_loop)

def wrap_periodic(xs, axes=()):
    """Return arrays wrapped around periodically

    Parameters
    ----------
    xs : array or sequence of arrays
    axes : sequence of integers, optional
        Extend arrays in xs by one in direction given by wrap
    """
    xs = xs if isinstance(xs, (list, tuple)) else [xs]
    ys = []
    for x in xs:
        if 0 in axes:
            x = np.vstack([x, x[0]])
        if 1 in axes:
            x = np.hstack([x, x[:, 0][:, None]])
        if 2 in axes:
            x = np.concatenate([x, x[:, :, 0][:, :, None]], axis=2)
        ys.append(x)
    if len(ys) == 1:
        return ys[0]
    return ys

def surf3D(u, mesh=None, backend='plotly', wrapaxes=(), slices=None, fig=None, **kw):
    """Plot surface embedded in 3D

    Parameters
    ----------
    u : Function or Array
    mesh : list, str or None, optional
        - list of Cartesian meshes [x, y, z]
        - 'quadrature' - use quadrature mesh
        - 'uniform' - use uniform mesh
    backend : str, optional
        plotly or mayavi
    wrapaxes : sequence of integers, optional
        For domains that wrap around, extend mesh/u by one in given direction
    slices : None or sequence of slices, optional
        If only part of u should be plotted
    fig : Figure instance, optional
    kw : Keyword arguments, optional
        Used by plotly Surface. Possibly colorscale.
    """
    from shenfun.forms.arguments import Function, Array

    if mesh is None:
        mesh = 'quadrature'

    if isinstance(mesh, str):
        assert isinstance(u, (Function, Array)), "u must be Function/Array if mesh is not given"
        T = u.function_space()
        x, y, z = T.local_cartesian_mesh(kind=mesh)
    else:
        x, y, z = mesh

    if isinstance(u, Function):
        u = u.backward(mesh=mesh)

    x, y, z, u = wrap_periodic([x, y, z, u], wrapaxes)

    if slices:
        x = x[slices]
        y = y[slices]
        z = z[slices]
        u = u[slices]

    if u.dtype.char in 'FDG':
        u = abs(u)

    if backend == 'plotly':
        import plotly.graph_objects as go
        s = go.Surface(x=x, y=y, z=z, surfacecolor=u, **kw)
        fig = go.Figure(s)
        d = {'visible': False, 'showgrid': False, 'zeroline': False}
        fig.update_layout(scene={'xaxis': d, 'yaxis': d, 'zaxis': d})

    elif backend == 'mayavi':
        from mayavi import mlab
        fig = mlab.figure(bgcolor=(1, 1, 1))
        mlab.mesh(x, y, z, scalars=u, colormap='jet')

    return fig

def quiver3D(u, mesh=None, wrapaxes=(), slices=None, fig=None, kind='quadrature', **kw):
    """

    Parameters
    ----------
    u : Array
    mesh : list of Cartesian meshes [x, y, z], optional
    backend : str, optional
        plotly or mayavi
    wrapaxes : sequence of integers, optional
        For domains that wrap around, extend mesh/u by one in given direction
    fig : None, optional
        If None create a new figure
    slices : None or sequence of slices, optional
        If only part of u should be plotted
    kind : str, optional
        - 'quadrature' - Use quadrature mesh
        - 'uniform' - Use uniform mesh
    kw : Keyword arguments, optional
        Arguments to mlab.quiver3d, for example 'scale_factor',
        'color' or ''mode

    """
    from shenfun.forms.arguments import Function, Array
    from mayavi import mlab

    par = {
        'scale_factor': 0.1,
        'color': (0, 0, 0),
        'mode': '2darrow',
    }
    par.update(kw)
    if mesh is None:
        assert isinstance(u, (Function, Array)), "u must be Function/Array if mesh is not given"
        T = u.function_space()
        x, y, z = T.local_cartesian_mesh(kind=kind)
    else:
        x, y, z = mesh

    if isinstance(u, Function):
        u = u.backward(mesh=kind)

    if u.dtype.char in 'FDG':
        u = abs(u)

    x, y, z = wrap_periodic([x, y, z], wrapaxes)
    u = wrap_periodic([u], np.array(wrapaxes)+1)
    if slices:
        x = x[slices]
        y = y[slices]
        z = z[slices]
        u = u[slices]
    if fig is None:
        mlab.figure(bgcolor=(1, 1, 1), size=(400, 400))
    mlab.quiver3d(x, y, z, u[0], u[1], u[2], **par)

def scalar_product(v, f, output_array=None, assemble='exact'):
    r"""Return scalar product

    .. math::

        (v, f)_w

    Parameters
    ----------
    v : :class:`.TestFunction`
    f : Sympy function
    output_array : :class:`.Function`
    assemble : str, optional

        - 'exact'
        - 'adaptive'

    Note
    ----
    The computed scalar product is not compensated for a non-standard domain size.

    >>> from shenfun import inner, FunctionSpace, TestFunction
    >>> import sympy as sp
    >>> C = FunctionSpace(4, 'C')
    >>> v = TestFunction(C)
    >>> x = sp.Symbol('x', real=True)
    >>> f = x**2
    >>> inner(v, f, assemble='exact')
    Function([1.57079633, 0.        , 0.78539816, 0.        ])

    """
    assert assemble in ('exact', 'adaptive')
    if assemble == 'adaptive':
        from scipy.integrate import quadrature
    T = v.function_space()
    if output_array is None:
        from shenfun import Function
        output_array = Function(T)

    if T.is_composite_space:
        for vi, xi in zip(v, output_array):
            xi = scalar_product(vi, f, xi)
        return output_array

    x = sp.Symbol('x', real=True)
    if not isinstance(f, Number):
        s = f.free_symbols
        assert len(s) == 1
        x = s.pop()
        xm = T.map_true_domain(x)
        if T.family() == 'chebyshev':
            f = f.subs(x, sp.cos(xm))
        else:
            f = f.subs(x, xm)

    if T.family() == 'chebyshev':
        S = T.stencil_matrix().diags('csr')
        for i in range(T.slice().start, T.slice().stop):
            M = S.getrow(i)
            integrand = sp.S(0)
            for ind, d in zip(M.indices, M.data):
                integrand += d*sp.cos(ind*x)
            integrand = f*integrand
            if assemble == 'exact':
                output_array[i] = sp.integrate(integrand, (x, (0, sp.pi)))
            elif assemble == 'adaptive':
                if isinstance(integrand, Number):
                    output_array[i] = integrand*np.pi
                else:
                    output_array[i] = quadrature(sp.lambdify(x, integrand), 0, np.pi)[0]
    else:
        domain = T.reference_domain()
        f *= T.weight()
        for i in range(T.slice().start, T.slice().stop):
            integrand = f*sp.conjugate(T.basis_function(i, x=x))
            if assemble == 'exact':
                output_array[i] = sp.integrate(integrand, (x, (domain[0], domain[1])))
            elif assemble == 'adaptive':
                if len(integrand.free_symbols) == 0:
                    output_array[i] = integrand*float(domain[1]-domain[0])
                else:
                    output_array[i] = quadrature(sp.lambdify(x, integrand), float(domain[0], float(domain[1])))
    if T.domain_factor() != 1:
        output_array /= float(T.domain_factor())
    return output_array



#pylint: disable=line-too-long, len-as-condition, missing-docstring, too-many-instance-attributes

import numpy as np
import scipy.linalg as scipy_la
from mpi4py import MPI

comm = MPI.COMM_WORLD

class Helmholtz_2dirichlet:
    """Helmholtz solver for 2-dimensional problems with 2 Dirichlet bases.

    .. math::

        a_0 BUB + a_1 AUB + a_2 BUA^T = F

    Somewhat experimental.

    """

    def __init__(self, matrices):

        self.V = np.zeros(0)
        self.matrices = matrices
        assert len(matrices) == 3

        # There are three terms, BUB, AUB and BUA
        # Extract A and B
        scale = {}
        for tmp in matrices:
            mats = tmp.mats
            if mats[0].get_key() == 'BSDSDmat' and mats[1].get_key() == 'BSDSDmat':
                B = mats[0]
                B1 = mats[1]
                scale['BUB'] = tmp.scale
                self.BB = tmp

            elif mats[0].get_key() == 'ASDSDmat' and mats[1].get_key() == 'BSDSDmat':
                A = mats[0]
                scale['AUB'] = tmp.scale
                self.AB = tmp

            else:
                A1 = mats[1]
                scale['BUA'] = tmp.scale
                self.BA = tmp

        # Create transfer object to realign data in second direction
        self.T = T = matrices[0].space
        pencilA = T.forward.output_pencil
        pencilB = T.forward.input_pencil
        self.pencilB = pencilB
        self.transAB = pencilA.transfer(pencilB, 'd')
        self.u_B = np.zeros(self.transAB.subshapeB)
        self.rhs_A = np.zeros(self.transAB.subshapeA)
        self.rhs_B = np.zeros(self.transAB.subshapeB)

        self.A = A
        self.B = B
        self.A1 = A1
        self.B1 = B1
        self.scale = scale

        self.lmbda = None
        self.lmbdax = None
        self.lmbday = None
        self.Vx = None
        self.Vy = None

    def solve_eigen_problem(self, A, B, solver):
        """Solve the eigen problem"""
        N = A.testfunction[0].N
        s = A.testfunction[0].slice()
        self.V = np.zeros((N, N))
        self.lmbda = np.ones(N)
        if solver == 0:
            self.lmbda[s], self.V[s, s] = scipy_la.eigh(A.diags().toarray(),
                                                        B.diags().toarray())

        elif solver == 1:
            #self.lmbda[s], self.V[s, s] = scipy_la.eigh(B.diags().toarray())
            a = np.zeros((3, N-2))
            a[0, :] = B[0]
            a[2, :-2] = B[2]
            self.lmbda[s], self.V[s, s] = scipy_la.eig_banded(a, lower=True)

    def __call__(self, b, u, solver=1):
        from shenfun import FunctionSpace, TensorProductSpace, TPMatrix, Identity, la

        if solver == 0: # pragma: no cover

            if len(self.V) == 0:
                self.solve_eigen_problem(self.A, self.B, solver)
                self.Vx = self.V
                self.lmbdax = self.lmbda
                if not self.A.testfunction[0].N == self.A1.testfunction[0].N:
                    self.Vx = self.V.copy()
                    self.lmbdax = self.lmbda.copy()
                    self.solve_eigen_problem(self.A1, self.B1, solver)
                    self.Vy = self.V
                    self.lmbday = self.lmbda
                else:
                    self.Vy = self.Vx
                    self.lmbday = self.lmbdax

            # Map the right hand side to eigen space
            u[:] = (self.Vx.T).dot(b)
            self.transAB.forward(u, self.u_B)
            self.u_B[:] = self.u_B.dot(self.Vy)
            self.transAB.backward(self.u_B, u)

            # Apply the inverse in eigen space
            ls = self.T.local_slice()
            u /= (self.BB.scale + self.lmbdax[:, np.newaxis] + self.lmbday[np.newaxis, :])[ls]

            # Map back to physical space
            u[:] = self.Vx.dot(u)
            self.transAB.forward(u, self.u_B)
            self.u_B[:] = self.u_B.dot(self.Vy.T)
            self.transAB.backward(self.u_B, u)

        elif solver == 1:
            #FIXME Should improve this after redesign of solvers

            if comm.Get_size() > 1:
                raise RuntimeError('Currently broken for MPI > 1')

            assert self.A.testfunction[0].is_scaled()

            if len(self.V) == 0:
                self.solve_eigen_problem(self.A, self.B, solver)

            ls = [slice(start, start+shape) for start, shape in zip(self.pencilB.substart,
                                                                    self.pencilB.subshape)]

            B1_scale = np.zeros((ls[0].stop-ls[0].start, 1))
            B1_scale[:, 0] = self.BB.scale + 1./self.lmbda[ls[0]]
            A1_scale = self.scale['AUB']
            # Create Helmholtz solver along axis=1
            N0 = self.matrices[0].mats[0].shape[0]

            F = FunctionSpace(N0, 'L', dtype='d')
            FT = TensorProductSpace(comm, (F, self.T.bases[1].get_unplanned()))
            AA = TPMatrix([Identity((N0, N0)), self.A1], FT, self.T, A1_scale)
            BB = TPMatrix([Identity((N0, N0)), self.B1], FT, self.T, B1_scale)
            AA._issimplified = True
            BB._issimplified = True
            AA.naxes = [1]
            BB.naxes = [1]

            Helmy = la.SolverGeneric1ND([AA, BB])
            # Map the right hand side to eigen space
            self.rhs_A = self.V.T.dot(b)
            self.rhs_A /= self.lmbda[:, np.newaxis]
            self.transAB.forward(self.rhs_A, self.rhs_B)
            self.u_B = Helmy(self.rhs_B, self.u_B, fast=False)
            self.transAB.backward(self.u_B, u)
            u[:] = self.V.dot(u)

        return u



import sympy as sp
import numpy as np
import scipy.special as sc
from scipy.special import jv
from shenfun.matrixbase import SparseMatrix
from shenfun.spectralbase import SpectralBase, getCompositeBase, getBCGeneric, BoundaryConditions

#pylint: disable=method-hidden,no-else-return,not-callable,abstract-method,no-member,cyclic-import

bases = ['Orthogonal',
         'CompactDirichlet']
bcbases = []
testbases = []
__all__ = bases + bcbases

xp = sp.Symbol('x', real=True)


class Orthogonal(SpectralBase):
    r"""Function space for a regular Bessel series

    The orthogonal basis is the Fourier-Bessel function of order 0

    .. math::

        \phi_k(x) = J_0(\lambda_k x/b), \quad k = 0, 1, \ldots, N-1,

    where :math:`\lambda_k` is the :math:`k`'th root of :math:`J_0(x)`
    and the domain is :math:`x\in [0, b]`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    Note
    ----
    We are using Fourier-Bessel functions and not the regular Bessel functions.

    """
    def __init__(self, N, dtype=float, padding_factor=1, dealias_direct=False,
                 domain=(0, 1), coordinates=None, **kw):
        SpectralBase.__init__(self, N, quad="BG", domain=domain, dtype=dtype,
                              padding_factor=padding_factor, dealias_direct=dealias_direct,
                              coordinates=coordinates)
        self.zeros = np.array([sc.jn_zeros(n, N) for n in range(N)])
        self.plan(int(N*padding_factor), 0, dtype, {})

    @staticmethod
    def family():
        return 'bessel'

    def reference_domain(self):
        return (0, 1)

    def points_and_weights(self, N=None, map_true_domain=False, weighted=True, **kw):
        if N is None:
            N = self.shape(False)
        assert self.quad == "BG"
        points = np.linspace(0, 1, self.N)
        weights = None

        if weighted:
            pass
        else:
            raise NotImplementedError

        return points, weights

    def vandermonde(self, x):
        if x is None:
            x = self.mesh(False, False)
        V = np.zeros((x.shape[0], self.N))
        for i in range(self.dim()):
            f = self.orthogonal_basis_function(i, x=xp)
            V[:, i] = sp.lambdify(xp, f)(x)
        return V

    @staticmethod
    def bnd_values(k=0, **kw):
        if k == 0:
            return (lambda i: int(i==0), None)
        elif k == 1:
            return (lambda i: 0, None)
        raise NotImplementedError

    def stencil_matrix(self, N=None):
        N = self.N if N is None else N
        return SparseMatrix({0: 1}, (N, N))

    def eval(self, x, u, output_array=None):
        x = np.atleast_1d(x)
        raise NotImplementedError

    def orthogonal_basis_function(self, i=0, x=xp):
        lambdai = self.zeros[i, -1]
        return sp.besselj(i, x*lambdai/self.domain[1])

    def L2_norm_sq(self, i):
        lambdai = self.zeros[i, -1]
        return sp.Rational(1, 2)*sp.besselj(i+1, lambdai)**2

    def l2_norm_sq(self, i=None):
        raise NotImplementedError

    def weight(self, x=xp):
        return x

    @property
    def is_orthogonal(self):
        return True

    def get_orthogonal(self, **kwargs):
        d = dict(quad=self.quad,
                 domain=self.domain,
                 dtype=self.dtype,
                 padding_factor=self.padding_factor,
                 dealias_direct=self.dealias_direct,
                 coordinates=self.coors.coordinates)
        d.update(kwargs)
        return Orthogonal(self.N, **d)

    @staticmethod
    def short_name():
        return 'Jn'

    #def get_bc_space(self):
    #    if self._bc_space:
    #        return self._bc_space
    #    self._bc_space = BCGeneric(self.N, bc=self.bcs)
    #    return self._bc_space

CompositeBase = getCompositeBase(Orthogonal)
#BCGeneric = getBCGeneric(CompositeBase)

class CompactDirichlet(CompositeBase):
    r"""Laguerre function space for Dirichlet boundary conditions

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::

        \phi_k &= (La_k - La_{k+1})\exp(-x/2), \, k=0, 1, \ldots, N-2, \\
        \phi_{N-1} &= L_0\exp(-x/2),

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(0) &= a

    The last basis function is for boundary condition and only used if a is
    different from 0. In one dimension :math:`\hat{u}_{N-1}=a`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Laguerre-Gauss
    bc : 1-tuple of number (a,)
        Boundary value at x=0
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, bc=(0,), dtype=float, padding_factor=1,
                 dealias_direct=False, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad='LG', dtype=dtype, padding_factor=padding_factor,
                               bc=bc, dealias_direct=dealias_direct, domain=(0, sp.S.Infinity),
                               coordinates=coordinates)
        self._stencil = {0: 1, 1: -1}

    @staticmethod
    def boundary_condition():
        return 'Dirichlet'

    @staticmethod
    def short_name():
        return 'SD'

#
#class CompactNeumann(CompositeBase):
#    r"""Laguerre function space for Dirichlet boundary conditions
#
#    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is
#
#    .. math::
#        \phi_k &= (La_k - \frac{2k+1}{2k+3}La_{k+1})\exp(-x/2), \, k=0, 1, \ldots, N-2, \\
#        \phi_{N-1} &= La_0\exp(-x/2),
#
#    such that
#
#    .. math::
#        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
#        u'(0) &= a
#
#    Parameters
#    ----------
#    N : int
#        Number of quadrature points
#    quad : str, optional
#        Type of quadrature
#
#        - LG - Laguerre-Gauss
#    bc : 1-tuple of number (a,)
#        Boundary value a = u'(0)
#    padding_factor : float, optional
#        Factor for padding backward transforms.
#    dealias_direct : bool, optional
#        Set upper 1/3 of coefficients to zero before backward transform
#    dtype : data-type, optional
#        Type of input data in real physical space. Will be overloaded when
#        basis is part of a :class:`.TensorProductSpace`.
#    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
#        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`
#
#    """
#    def __init__(self, N, bc=(0,), dtype=float, padding_factor=1,
#                 dealias_direct=False, coordinates=None, **kw):
#        if isinstance(bc, (tuple, list)):
#            assert len(bc) == 1
#            bc = BoundaryConditions({'left': {'N': bc[0]}, 'right': {}}, domain=(0, np.inf))
#        CompositeBase.__init__(self, N, dtype=dtype, quad='LG', padding_factor=padding_factor,
#                               bc=bc, dealias_direct=dealias_direct, domain=(0, sp.S.Infinity),
#                               coordinates=coordinates)
#        self._stencil = {0: 1, 1: -(2*n+1)/(2*n+3)}
#
#    @staticmethod
#    def boundary_condition():
#        return 'Neumann'
#
#    @staticmethod
#    def short_name():
#        return 'CN'
#
#
#class Generic(CompositeBase):
#    r"""Function space for Laguerre space with any boundary conditions
#
#    Any combination of Dirichlet and Neumann is possible.
#
#    Parameters
#    ----------
#    N : int, optional
#        Number of quadrature points
#    quad : str, optional
#        Type of quadrature
#
#        - LG - Laguerre-Gauss
#
#    bc : dict, optional
#        The dictionary must have key 'left' (not 'right'), to describe boundary
#        conditions on the left boundary. Specify Dirichlet with
#
#            {'left': {'D': a}}
#
#        for some value `a`, that will be neglected in the current
#        function. Specify mixed Neumann and Dirichlet as
#
#            {'left': {'D': a, 'N': b}}
#
#        See :class:`~shenfun.spectralbase.BoundaryConditions`.
#    domain : 2-tuple of floats, optional
#        The computational domain
#    dtype : data-type, optional
#        Type of input data in real physical space. Will be overloaded when
#        basis is part of a :class:`.TensorProductSpace`.
#    padding_factor : float, optional
#        Factor for padding backward transforms.
#    dealias_direct : bool, optional
#        Set upper 1/3 of coefficients to zero before backward transform
#    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
#        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`
#
#    """
#    def __init__(self, N, quad="LG", bc={}, domain=(0, sp.S.Infinity), dtype=float,
#                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
#        from shenfun.utilities.findbasis import get_stencil_matrix
#        self._stencil = get_stencil_matrix(bc, 'laguerre')
#        if not isinstance(bc, BoundaryConditions):
#            bc = BoundaryConditions(bc, domain=domain)
#        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
#                               padding_factor=padding_factor, dealias_direct=dealias_direct,
#                               coordinates=coordinates)
#
#    @staticmethod
#    def boundary_condition():
#        return 'Generic'
#
#    @staticmethod
#    def short_name():
#        return 'GL'



r"""
This module contains specific inner product matrices for the different bases in
the Chebyshev family of the second kind.

A naming convention is used for the first capital letter for all matrices.
The first letter refers to type of matrix.

    - Mass matrices start with `B`
    - One derivative start with `C`
    - Two derivatives (Laplace) start with `A`
    - Four derivatives (Biharmonic) start with `S`

A matrix may consist of different types of test and trialfunctions as long as
they are all in the Chebyshev family, either first or second kind.
The next letters in the matrix name uses the 'short_name' for all these
different bases, see chebyshevu.bases.py.

All matrices in this module may be looked up using the 'mat' dictionary,
which takes test and trialfunctions along with the number of derivatives
to be applied to each. As such the mass matrix BTTmat may be looked up
as

>>> from shenfun.chebyshevu.matrices import mat
>>> from shenfun.chebyshevu.bases import Orthogonal as T
>>> B = mat[((T, 0), (T, 0))]

and an instance of the matrix can be created as

>>> B0 = T(10)
>>> BM = B((B0, 0), (B0, 0))
>>> import numpy as np
>>> d = {0: np.pi/2}
>>> [np.all(BM[k] == v) for k, v in d.items()]
[True]

However, this way of creating matrices is not reccommended use. It is far
more elegant to use the TrialFunction/TestFunction interface, and to
generate the matrix as an inner product:

>>> from shenfun import TrialFunction, TestFunction, inner
>>> u = TrialFunction(B0)
>>> v = TestFunction(B0)
>>> BM = inner(u, v)
>>> [np.all(BM[k] == v) for k, v in d.items()]
[True]

To see that this is in fact the BUUmat:

>>> print(BM.__class__)
<class 'shenfun.chebyshevu.matrices.BUUmat'>

"""
#pylint: disable=bad-continuation, redefined-builtin

from __future__ import division

import numpy as np
from shenfun.matrixbase import SpectralMatrix, SparseMatrix, SpectralMatDict
from shenfun.spectralbase import get_norm_sq
from shenfun.la import TwoDMA
from shenfun.chebyshev import bases as chebbases
from . import bases

# Short names for instances of bases
U = bases.Orthogonal
P1 = bases.Phi1

SD = chebbases.ShenDirichlet
SN = chebbases.ShenNeumann


class BUUmat(SpectralMatrix):
    r"""Mass matrix :math:`B=(b_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        b_{kj}=(U_j, U_k)_w,

    :math:`U_k \in` :class:`.chebyshevu.bases.Orthogonal` and test and trial spaces have
    dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test, trial = self.testfunction, self.trialfunction
        assert isinstance(test[0], U)
        assert isinstance(trial[0], U)
        return {0: get_norm_sq(test[0], trial[0], method)}

class BP1SDmat(SpectralMatrix):
    r"""Mass matrix :math:`B=(b_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        b_{kj}=(\psi_j, \phi_k)_w,

    where the test function :math:`\phi_k \in` :class:`.chebyshevu.bases.Phi1`, the
    trial :math:`\psi_j \in` :class:`.chebyshev.bases.ShenDirichlet`, and test and
    trial spaces have dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test, trial = self.testfunction, self.trialfunction
        from shenfun.jacobi.recursions import Lmat, half, cn
        assert isinstance(test[0], P1)
        assert isinstance(trial[0], SD)
        assert test[0].quad == 'GU'
        N = test[0].N-2
        K = trial[0].stencil_matrix()
        K.shape = (N, N+2)
        B2 = Lmat(2, 0, 2, N, N+2, -half, -half, cn) # B^{(2)_{(2)}}
        if not test[0].is_scaled():
            k = np.arange(N+2)
            B2 = SparseMatrix({0: (k[:-2]+2)}, (N, N)).diags('csr')*B2
        M = B2 * K.diags('csr').T
        K.shape = (N+2, N+2)
        d = {-2: M.diagonal(-2), 0: M.diagonal(0), 2: M.diagonal(2), 4: M.diagonal(4)}
        return d

class AP1SDmat(SpectralMatrix):
    r"""Stiffness matrix :math:`A=(a_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        a_{kj}=(\psi''_j, \phi_k)_w,

    where the test function :math:`\phi_k \in` :class:`.chebyshevu.bases.Phi1`, the trial
    function :math:`\psi_j \in` :class:`.chebyshev.bases.ShenDirichlet`, and test and
    trial spaces have dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test, trial = self.testfunction, self.trialfunction
        assert isinstance(test[0], P1)
        assert isinstance(trial[0], SD)
        d = {0: -1, 2: 1}
        if not test[0].is_scaled():
            k = np.arange(test[0].N-2)
            d = {0: -(k+2), 2: k[:-2]+2}
        return d

    def get_solver(self):
        return TwoDMA

class BP1SNmat(SpectralMatrix):
    r"""Mass matrix :math:`B=(b_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        b_{kj}=(\psi_j, \phi_k)_w,

    where the test function :math:`\phi_k \in` :class:`.chebyshevu.bases.Phi1`, the
    trial function :math:`\psi_j \in` :class:`.chebyshev.bases.ShenNeumann`, and test and
    trial spaces have dimensions of M and N, respectively.

    """
    def assemble(self, method):
        from shenfun.jacobi.recursions import Lmat, half, cn
        test, trial = self.testfunction, self.trialfunction
        assert isinstance(test[0], P1)
        assert isinstance(trial[0], SN)
        N = test[0].N-2
        K = trial[0].stencil_matrix()
        K.shape = (N, N+2)
        B2 = Lmat(2, 0, 2, N, N+2, -half, -half, cn) # B^{(2)_{(2)}}
        if not test[0].is_scaled():
            k = np.arange(test[0].N)
            B2 = SparseMatrix({0: (k[:-2]+2)}, (N, N)).diags('csr')*B2
        M = B2 * K.diags('csr').T
        K.shape = (N+2, N+2)
        d = {-2: M.diagonal(-2), 0: M.diagonal(0), 2: M.diagonal(2), 4: M.diagonal(4)}
        return d

class AP1SNmat(SpectralMatrix):
    r"""Stiffness matrix :math:`A=(a_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        a_{kj}=(\psi''_j, \phi_k)_w,

    where the test function :math:`\phi_k \in` :class:`.chebyshevu.bases.Phi1`, the trial
    function :math:`\psi_j \in` :class:`.chebyshev.bases.ShenNeumann`, and test and
    trial spaces have dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test, trial = self.testfunction, self.trialfunction
        assert isinstance(test[0], P1)
        assert isinstance(trial[0], SN)
        k = np.arange(test[0].N-2)
        d = {0: -(k/(k+2))**2, 2: 1}
        if not test[0].is_scaled():
            d = {0: -k**2/(k+2), 2: k[:-2]+2}
        return d

    def get_solver(self):
        return TwoDMA

# Define dictionary to hold all predefined matrices
# When looked up, missing matrices will be generated automatically
mat = SpectralMatDict({
    ((U,  0), (U,  0)): BUUmat,
    ((P1, 0), (SD, 0)): BP1SDmat,
    ((P1, 0), (SN, 0)): BP1SNmat,
    ((P1, 0), (SD, 2)): AP1SDmat,
    ((P1, 0), (SN, 2)): AP1SNmat,
})



from copy import copy, deepcopy
import functools
from ipaddress import ip_address
from itertools import product
import numpy as np
import sympy as sp
from mpi4py import MPI
import pytest
import mpi4py_fft
import shenfun
from shenfun.chebyshev import matrices as cmatrices
from shenfun.chebyshev import bases as cbases
from shenfun.chebyshevu import matrices as cumatrices
from shenfun.chebyshevu import bases as cubases
from shenfun.ultraspherical import matrices as umatrices
from shenfun.ultraspherical import bases as ubases
from shenfun.legendre import matrices as lmatrices
from shenfun.legendre import bases as lbases
from shenfun.laguerre import matrices as lagmatrices
from shenfun.laguerre import bases as lagbases
from shenfun.hermite import matrices as hmatrices
from shenfun.hermite import bases as hbases
from shenfun.jacobi import matrices as jmatrices
from shenfun.jacobi import bases as jbases
from shenfun.chebyshev import la as cla
from shenfun import div, grad, inner, TensorProductSpace, FunctionSpace, SparseMatrix,\
    Function, comm, VectorSpace, BlockMatrix, CompositeSpace
from shenfun.spectralbase import inner_product
from shenfun.config import config

x = sp.symbols('x', real=True)
xp = sp.symbols('x', real=True, positive=True)

ctrialBasis = [cbases.__dict__.get(base) for base in cbases.bases[:-1]]
ctestBasis = ctrialBasis + [cbases.__dict__.get(base) for base in cbases.testbases[:-1]]
cutrialBasis = [cubases.__dict__.get(base) for base in cubases.bases[:-1]]
cutestBasis = cutrialBasis + [cubases.__dict__.get(base) for base in cubases.testbases[:-1]]

utrialBasis = [ubases.__dict__.get(base) for base in ubases.bases[:-1]]
utestBasis = utrialBasis + [ubases.__dict__.get(base) for base in ubases.testbases[:-1]]
ltrialBasis = [lbases.__dict__.get(base) for base in lbases.bases[:-1]]
ltestBasis = ltrialBasis + [lbases.__dict__.get(base) for base in lbases.testbases[:-1]]
latrialBasis = [lagbases.__dict__.get(base) for base in lagbases.bases[:-1]]
htrialBasis = [hbases.__dict__.get(base) for base in hbases.bases]
jtrialBasis = [jbases.__dict__.get(base) for base in jbases.bases[:-1]]
jtestBasis = jtrialBasis + [jbases.__dict__.get(base) for base in jbases.testbases]

bcbases = (
    cbases.BCGeneric,
    lbases.BCGeneric,
    cubases.BCGeneric,
    ubases.BCGeneric
)

cquads = ('GC', 'GL')
cuquads = ('GU', )
uquads = ('QG',)
lquads = ('LG', 'GL')
lagquads = ('LG',)
hquads = ('HG',)
jquads = ('JG',)

bcs = {
    0: {'left': {'D': 1}, 'right': {'D': 1}},
    1: {'left': {'N': 1}, 'right': {'N': 1}},
    2: {'left': {'D': 1}, 'right': {'N': 1}},
    3: {'left': {'N': 1}, 'right': {'D': 1}},
    4: {'left': {'D': 1, 'N': 1}},
    5: {'right': {'D': 1, 'N': 1}}
}

for f in ['dct', 'dst', 'fft', 'ifft', 'rfft', 'irfft']:
    config['fftw'][f]['planner_effort'] = 'FFTW_ESTIMATE'

N = 14
k = np.arange(N).astype(float)
a = np.random.random(N)
c = np.zeros(N)
c1 = np.zeros(N)

work = {
    3:
    (np.random.random((N, N, N)),
     np.zeros((N, N, N)),
     np.zeros((N, N, N))),
    2:
    (np.random.random((N, N)),
     np.zeros((N, N)),
     np.zeros((N, N)))
    }

cbases2 = list(product(ctestBasis, ctrialBasis))
cubases2 = list(product(cutestBasis, cutrialBasis))
ubases2 = list(product(utestBasis, utrialBasis))
lbases2 = list(product(ltestBasis, ltrialBasis))
lagbases2 = list(product(latrialBasis, latrialBasis))
hbases2 = list(product(htrialBasis, htrialBasis))
jbases2 = list(product(jtestBasis, jtrialBasis))
bases2 = cbases2+lbases2+lagbases2+hbases2+jbases2+cubases2+ubases2

cmats_and_quads = [list(k[0])+[k[1]] for k in product([(k, v) for k, v in cmatrices.mat.items()], cquads)]
cumats_and_quads = [list(k[0])+[k[1]] for k in product([(k, v) for k, v in cumatrices.mat.items()], cuquads)]
umats_and_quads = [list(k[0])+[k[1]] for k in product([(k, v) for k, v in umatrices.mat.items()], uquads)]
lmats_and_quads = [list(k[0])+[k[1]] for k in product([(k, v) for k, v in lmatrices.mat.items()], lquads)]
lagmats_and_quads = [list(k[0])+[k[1]] for k in product([(k, v) for k, v in lagmatrices.mat.items()], lagquads)]
hmats_and_quads = [list(k[0])+[k[1]] for k in product([(k, v) for k, v in hmatrices.mat.items()], hquads)]
jmats_and_quads = [list(k[0])+[k[1]] for k in product([(k, v) for k, v in jmatrices.mat.items()], jquads)]
mats_and_quads = cmats_and_quads+lmats_and_quads+lagmats_and_quads+hmats_and_quads+jmats_and_quads+umats_and_quads+cumats_and_quads

#cmats_and_quads_ids = ['-'.join(i) for i in product([v.__name__ for v in cmatrices.mat.values()], cquads)]
#lmats_and_quads_ids = ['-'.join(i) for i in product([v.__name__ for v in lmatrices.mat.values()], lquads)]

num_tests = 10
some_mats_and_quads = [mats_and_quads[i] for i in np.random.randint(0, len(mats_and_quads), num_tests)]
some_cbases2 = [cbases2[i] for i in np.random.randint(0, len(cbases2), num_tests)]
some_cubases2 = [cubases2[i] for i in np.random.randint(0, len(cubases2), num_tests)]
some_lbases2 = [lbases2[i] for i in np.random.randint(0, len(lbases2), num_tests)]
some_jbases2 = [jbases2[i] for i in np.random.randint(0, len(jbases2), num_tests)]


@pytest.mark.parametrize('key, mat, quad', mats_and_quads)
def test_mat(key, mat, quad):
    """Test that matrices built by hand equals those automatically generated"""
    test = key[0]
    trial = key[1]
    measure = 1
    if len(key) == 3:
        measure = key[2]

    t0 = test[0]
    t1 = trial[0]

    if trial[0] in bcbases:
        # Just use some random boundary condition
        t1 = functools.partial(t1, bc=bcs[np.random.randint(0, 6)])

    testfunction = (t0(N, quad=quad), test[1])
    trialfunction = (t1(N, quad=quad), trial[1])
    mat = mat(testfunction, trialfunction, measure=measure)
    fixed_resolution = None if sp.degree(sp.sympify(measure)) < 2 else int(3*N/2)
    shenfun.check_sanity(mat, testfunction, trialfunction, measure, fixed_resolution=fixed_resolution)
    if test[0].family() == 'Legendre' and test[0].boundary_condition() == 'Dirichlet':
        testfunction = (test[0](N, quad=quad, scaled=True), test[1])
        trialfunction = (trial[0](N, quad=quad, scaled=True), trial[1])
        mat = mat(testfunction, trialfunction)
        shenfun.check_sanity(mat, testfunction, trialfunction, measure)


@pytest.mark.parametrize('b0,b1', some_cbases2)
@pytest.mark.parametrize('quad', cquads)
@pytest.mark.parametrize('k', range(5))
def test_cmatvec(b0, b1, quad, k):
    """Test matrix-vector product"""
    global c, c1
    b0 = b0(N, quad=quad)
    b1 = b1(N, quad=quad)
    mat = inner_product((b0, 0), (b1, k))
    formats = mat._matvec_methods + ['python', 'csr']
    c = mat.matvec(a, c, format='csr')
    for format in formats:
        c1 = mat.matvec(a, c1, format=format)
        assert np.allclose(c, c1)
    for dim in (2, 3):
        b, d, d1 = work[dim]
        for axis in range(0, dim):
            d = mat.matvec(b, d, format='csr', axis=axis)
            for format in formats:
                d1 = mat.matvec(b, d1, format=format, axis=axis)
                assert np.allclose(d, d1)

@pytest.mark.parametrize('b0,b1', some_cubases2)
@pytest.mark.parametrize('quad', cuquads)
@pytest.mark.parametrize('k', range(5))
def test_cumatvec(b0, b1, quad, k):
    """Test matrix-vector product"""
    global c, c1
    b0 = b0(N, quad=quad)
    b1 = b1(N, quad=quad)
    mat = inner_product((b0, 0), (b1, k))
    formats = mat._matvec_methods + ['python', 'csr']
    c = mat.matvec(a, c, format='csr')
    for format in formats:
        c1 = mat.matvec(a, c1, format=format)
        assert np.allclose(c, c1)
    for dim in (2, 3):
        b, d, d1 = work[dim]
        for axis in range(0, dim):
            d = mat.matvec(b, d, format='csr', axis=axis)
            for format in formats:
                d1 = mat.matvec(b, d1, format=format, axis=axis)
                assert np.allclose(d, d1)

@pytest.mark.parametrize('b0,b1', some_lbases2)
@pytest.mark.parametrize('quad', ('LG',))
@pytest.mark.parametrize('k0,k1', product((0, 1, 2), (0, 1, 2)))
def test_lmatvec(b0, b1, quad, k0, k1):
    """Test matrix-vector product"""
    global c, c1, a
    b0 = b0(N, quad=quad)
    b1 = b1(N, quad=quad)
    mat = inner_product((b0, k0), (b1, k1))
    c = mat.matvec(a, c, format='dia')
    formats = mat._matvec_methods + ['python', 'csr']
    for format in formats:
        c1 = mat.matvec(a, c1, format=format)
        assert np.allclose(c, c1)

    for dim in (2, 3):
        b, d, d1 = work[dim]
        for axis in range(0, dim):
            d = mat.matvec(b, d, format='dia', axis=axis)
            for formats in formats:
                d1 = mat.matvec(b, d1, format=format, axis=axis)
                assert np.allclose(d, d1)


@pytest.mark.parametrize('b0,b1', lagbases2)
@pytest.mark.parametrize('quad', lagquads)
@pytest.mark.parametrize('format', ('dia', 'python'))
@pytest.mark.parametrize('k0,k1', product((0, 1, 2), (0, 1, 2)))
def test_lagmatvec(b0, b1, quad, format, k0, k1):
    """Test matrix-vector product"""
    global c, c1
    b0 = b0(N, quad=quad)
    b1 = b1(N, quad=quad)
    mat = inner_product((b0, k0), (b1, k1))
    c = mat.matvec(a, c, format='csr')
    formats = mat._matvec_methods + ['python', 'csr']
    for format in formats:
        c1 = mat.matvec(a, c1, format=format)
        assert np.allclose(c, c1)
    for dim in (2, 3):
        b, d, d1 = work[dim]
        for axis in range(0, dim):
            d = mat.matvec(b, d, format='csr', axis=axis)
            for format in formats:
                d1 = mat.matvec(b, d1, format=format, axis=axis)
                assert np.allclose(d, d1)

@pytest.mark.parametrize('b0,b1', hbases2)
@pytest.mark.parametrize('quad', hquads)
@pytest.mark.parametrize('format', ('dia', 'python'))
@pytest.mark.parametrize('k0,k1', product((0, 1, 2), (0, 1, 2)))
def test_hmatvec(b0, b1, quad, format, k0, k1):
    """Test matrix-vector product"""
    global c, c1
    b0 = b0(N, quad=quad)
    b1 = b1(N, quad=quad)
    mat = inner_product((b0, k0), (b1, k1))
    formats = mat._matvec_methods + ['python', 'csr']
    c = mat.matvec(a, c, format='csr')
    for format in formats:
        c1 = mat.matvec(a, c1, format=format)
        assert np.allclose(c, c1)
    for dim in (2, 3):
        b, d, d1 = work[dim]
        for axis in range(0, dim):
            d = mat.matvec(b, d, format='csr', axis=axis)
            for format in formats:
                d1 = mat.matvec(b, d1, format=format, axis=axis)
                assert np.allclose(d, d1)

@pytest.mark.parametrize('b0,b1', some_jbases2)
@pytest.mark.parametrize('quad', jquads)
@pytest.mark.parametrize('format', ('dia', 'python'))
@pytest.mark.parametrize('k0,k1', product((0, 1, 2), (0, 1, 2)))
def test_jmatvec(b0, b1, quad, format, k0, k1):
    """Test matrix-vector product"""
    global c, c1
    b0 = b0(N, quad=quad)
    b1 = b1(N, quad=quad)

    mat = inner_product((b0, k0), (b1, k1))
    c = mat.matvec(a, c, format='csr')
    formats = mat._matvec_methods + ['python', 'csr']
    for format in formats:
        c1 = mat.matvec(a, c1, format=format)
        assert np.allclose(c, c1)
    dim = 2
    b, d, d1 = work[dim]
    for axis in range(0, dim):
        d = mat.matvec(b, d, format='csr', axis=axis)
        for format in formats:
            d1 = mat.matvec(b, d1, format=format, axis=axis)
            assert np.allclose(d, d1)

def test_eq():
    m0 = SparseMatrix({0: 1, 2: 2}, (6, 6))
    m1 = SparseMatrix({0: 1., 2: 2.}, (6, 6))
    assert m0 == m1
    assert m0 is not m1
    m2 = SparseMatrix({0: 1., 2: 3.}, (6, 6))
    assert m0 != m2

@pytest.mark.parametrize('key, mat, quad', some_mats_and_quads)
def test_imul(key, mat, quad):
    test = key[0]
    trial = key[1]
    measure = 1
    if len(key) == 3:
        measure = key[2]
        if quad == 'GL':
            return

    t0 = test[0]
    t1 = trial[0]
    if trial[0] in bcbases:
        # Just use some random boundary condition
        t1 = functools.partial(t1, bc=bcs[np.random.randint(0, 6)])
    testfunction = (t0(N, quad=quad), test[1])
    trialfunction = (t1(N, quad=quad), trial[1])
    try:
        mat = mat(testfunction, trialfunction, measure=measure)
    except AssertionError:
        return
    mc = mat.scale
    mat *= 2
    assert mat.scale == 2.0*mc

    mat.scale = mc
    mat = SparseMatrix(copy(dict(mat)), mat.shape)
    mat *= 2
    assert mat.scale == 2.0

@pytest.mark.parametrize('key, mat, quad', some_mats_and_quads)
def test_mul(key, mat, quad):
    test = key[0]
    trial = key[1]
    measure = 1
    if len(key) == 3:
        measure = key[2]
        if quad == 'GL':
            return

    t0 = test[0]
    t1 = trial[0]
    if trial[0] in bcbases:
        # Just use some random boundary condition
        t1 = functools.partial(t1, bc=bcs[np.random.randint(0, 6)])
    testfunction = (t0(N, quad=quad), test[1])
    trialfunction = (t1(N, quad=quad), trial[1])
    try:
        m = mat(testfunction, trialfunction, measure=measure)
    except AssertionError:
        return
    mc = 2.*m
    assert mc.scale == 2.0*m.scale

    mat = SparseMatrix(copy(dict(m)), m.shape)
    mc = 2.*mat
    assert mc.scale == 2.0

def test_mul2():
    mat = SparseMatrix({0: 1}, (3, 3))
    v = np.ones(3)
    c = mat * v
    assert np.allclose(c, 1)
    mat = SparseMatrix({-2:1, -1:1, 0: 1, 1:1, 2:1}, (3, 3))
    c = mat * v
    assert np.allclose(c, 3)
    SD = FunctionSpace(8, "L", bc=(0, 0), scaled=True)
    u = shenfun.TrialFunction(SD)
    v = shenfun.TestFunction(SD)
    mat = inner(grad(u), grad(v))
    z = Function(SD, val=1)
    c = mat * z
    assert np.allclose(c[:6], 1)

@pytest.mark.parametrize('key, mat, quad', some_mats_and_quads)
def test_rmul(key, mat, quad):
    test = key[0]
    trial = key[1]
    measure = 1
    if len(key) == 3:
        measure = key[2]
        if quad == 'GL':
            return

    t0 = test[0]
    t1 = trial[0]
    if trial[0] in bcbases:
        # Just use some random boundary condition
        t1 = functools.partial(t1, bc=bcs[np.random.randint(0, 6)])
    testfunction = (t0(N, quad=quad), test[1])
    trialfunction = (t1(N, quad=quad), trial[1])
    try:
        m = mat(testfunction, trialfunction, measure=measure)
    except AssertionError:
        return
    mc = m*2.
    assert mc.scale == 2.0*m.scale

    mat = SparseMatrix(copy(dict(m)), m.shape)
    mc = mat*2.
    assert mc.scale == 2.0

@pytest.mark.parametrize('key, mat, quad', some_mats_and_quads)
def test_div(key, mat, quad):
    test = key[0]
    trial = key[1]
    measure = 1
    if len(key) == 3:
        measure = key[2]
        if quad == 'GL':
            return

    t0 = test[0]
    t1 = trial[0]
    if trial[0] in bcbases:
        # Just use some random boundary condition
        t1 = functools.partial(t1, bc=bcs[np.random.randint(0, 6)])
    testfunction = (t0(N, quad=quad), test[1])
    trialfunction = (t1(N, quad=quad), trial[1])
    try:
        m = mat(testfunction, trialfunction, measure=measure)
    except AssertionError:
        return

    mc = m/2.
    assert mc.scale == 0.5*m.scale

    mat = SparseMatrix(copy(dict(m)), m.shape)
    mc = mat/2.
    assert mc.scale == 0.5

@pytest.mark.parametrize('basis, quad', list(product(ctrialBasis, cquads))+
                         list(product(ltrialBasis, lquads))+list(product(latrialBasis, lagquads)))
def test_div2(basis, quad):
    B = basis(8, quad=quad)
    u = shenfun.TrialFunction(B)
    v = shenfun.TestFunction(B)
    m = inner(u, v)
    z = Function(B, val=1)
    c = m / z
    #m2 = m.diags('csr')
    #c2 = spsolve(m2, z[B.slice()])

    c2 = Function(B)
    c2 = m.solve(z, c2)
    assert np.allclose(c2[B.slice()], c[B.slice()])

@pytest.mark.parametrize('key, mat, quad', some_mats_and_quads)
def test_add(key, mat, quad):
    test = key[0]
    trial = key[1]
    measure = 1
    if len(key) == 3:
        measure = key[2]
        if quad == 'GL':
            return

    t0 = test[0]
    t1 = trial[0]
    if trial[0] in bcbases:
        # Just use some random boundary condition
        t1 = functools.partial(t1, bc=bcs[np.random.randint(0, 6)])
    testfunction = (t0(N, quad=quad), test[1])
    trialfunction = (t1(N, quad=quad), trial[1])
    try:
        m = mat(testfunction, trialfunction, measure=measure)
    except AssertionError:
        return

    mc = m + m
    assert np.linalg.norm(mc.diags('csr').data-m.diags('csr').data*2) < 1e-8

    mat = SparseMatrix(copy(dict(m)), m.shape, m.scale)
    mc = m + mat
    assert np.linalg.norm(mc.diags('csr').data-m.diags('csr').data*2) < 1e-8

@pytest.mark.parametrize('key, mat, quad', some_mats_and_quads)
def test_iadd(key, mat, quad):
    test = key[0]
    trial = key[1]
    measure = 1
    if len(key) == 3:
        measure = key[2]
        if quad == 'GL':
            return

    t0 = test[0]
    t1 = trial[0]
    if trial[0] in bcbases:
        # Just use some random boundary condition
        t1 = functools.partial(t1, bc=bcs[np.random.randint(0, 6)])
    testfunction = (t0(N, quad=quad), test[1])
    trialfunction = (t1(N, quad=quad), trial[1])
    try:
        m = mat(testfunction, trialfunction, measure=measure)
    except AssertionError:
        return
    mc = m.copy()
    m += mc
    assert np.linalg.norm(m.diags('csr').data-mc.diags('csr').data*2) < 1e-8
    m -= 2*mc
    assert np.linalg.norm(m.diags('csr').data) < 1e-8

@pytest.mark.parametrize('key, mat, quad', some_mats_and_quads)
def test_isub(key, mat, quad):
    test = key[0]
    trial = key[1]
    measure = 1
    if len(key) == 3:
        measure = key[2]
        if quad == 'GL':
            return
    t0 = test[0]
    t1 = trial[0]
    if trial[0] in bcbases:
        # Just use some random boundary condition
        t1 = functools.partial(t1, bc=bcs[np.random.randint(0, 6)])
    testfunction = (t0(N, quad=quad), test[1])
    trialfunction = (t1(N, quad=quad), trial[1])
    try:
        m = mat(testfunction, trialfunction, measure=measure)
    except AssertionError:
        return
    mc = m.copy()
    m -= mc
    assert np.linalg.norm(m.diags('csr').data) < 1e-8

    m1 = SparseMatrix(deepcopy(dict(mc)), m.shape)
    m2 = SparseMatrix(deepcopy(dict(mc)), m.shape)
    m1 -= m2
    assert np.linalg.norm(m1.diags('csr').data) < 1e-8

@pytest.mark.parametrize('key, mat, quad', some_mats_and_quads)
def test_sub(key, mat, quad):
    test = key[0]
    trial = key[1]
    measure = 1
    if len(key) == 3:
        measure = key[2]
        if quad == 'GL':
            return
    t0 = test[0]
    t1 = trial[0]
    if trial[0] in bcbases:
        # Just use some random boundary condition
        t1 = functools.partial(t1, bc=bcs[np.random.randint(0, 6)])
    testfunction = (t0(N, quad=quad), test[1])
    trialfunction = (t1(N, quad=quad), trial[1])
    try:
        m = mat(testfunction, trialfunction, measure=measure)
    except AssertionError:
        return
    mc = m - m
    assert np.linalg.norm(mc.diags('csr').data) < 1e-8

    m1 = SparseMatrix(copy(dict(m)), m.shape)
    m2 = SparseMatrix(copy(dict(m)), m.shape)

    mc = m1 - m2
    assert np.linalg.norm(mc.diags('csr').data) < 1e-8

allaxes2D = {0: (0, 1), 1: (1, 0)}
allaxes3D = {0: (0, 1, 2), 1: (1, 0, 2), 2: (2, 0, 1)}

@pytest.mark.parametrize('axis', (0, 1, 2))
@pytest.mark.parametrize('family', ('chebyshev',))
def test_helmholtz3D(family, axis):
    la = cla
    N = (8, 9, 10)
    SD = FunctionSpace(N[allaxes3D[axis][0]], family=family, bc=(0, 0))
    K1 = FunctionSpace(N[allaxes3D[axis][1]], family='F', dtype='D')
    K2 = FunctionSpace(N[allaxes3D[axis][2]], family='F', dtype='d')
    subcomms = mpi4py_fft.pencil.Subcomm(MPI.COMM_WORLD, [0, 1, 1])
    bases = [0]*3
    bases[allaxes3D[axis][0]] = SD
    bases[allaxes3D[axis][1]] = K1
    bases[allaxes3D[axis][2]] = K2
    T = TensorProductSpace(subcomms, bases, axes=allaxes3D[axis], modify_spaces_inplace=True)
    u = shenfun.TrialFunction(T)
    v = shenfun.TestFunction(T)
    mat = inner(v, div(grad(u)))
    G = shenfun.la.Solver3D(mat)

    H = la.Helmholtz(*mat)
    uh = Function(T)
    s = SD.sl[SD.slice()]
    uh[s] = np.random.random(uh[s].shape) + 1j*np.random.random(uh[s].shape)
    f = Function(T)
    f = H.matvec(uh, f)

    gg = Function(T)
    gg = G.matvec(uh, gg)
    assert np.linalg.norm(f-gg) < 1e-12, np.linalg.norm(f-gg)

    uc = Function(T)
    uc = H(f, uc)
    assert np.linalg.norm(uc-uh) < 1e-12
    T.destroy()

@pytest.mark.parametrize('axis', (0, 1))
@pytest.mark.parametrize('family', ('chebyshev',))
def test_helmholtz2D(family, axis):
    la = cla
    N = (8, 9)
    SD = FunctionSpace(N[axis], family=family, bc=(0, 0))
    K1 = FunctionSpace(N[(axis+1)%2], family='F', dtype='d')
    subcomms = mpi4py_fft.pencil.Subcomm(MPI.COMM_WORLD, allaxes2D[axis])
    bases = [K1]
    bases.insert(axis, SD)
    T = TensorProductSpace(subcomms, bases, axes=allaxes2D[axis], modify_spaces_inplace=True)
    u = shenfun.TrialFunction(T)
    v = shenfun.TestFunction(T)
    mat = inner(v, div(grad(u)))
    H = la.Helmholtz(*mat)
    G = shenfun.la.Solver2D(mat)
    uh = Function(T)
    s = SD.sl[SD.slice()]
    uh[s] = np.random.random(uh[s].shape) + 1j*np.random.random(uh[s].shape)
    f = Function(T)
    f = H.matvec(uh, f)
    gg = Function(T)
    gg = G.matvec(uh, gg)
    assert np.linalg.norm(f-gg) < 1e-12, np.linalg.norm(f-gg)
    uc = Function(T)
    uc = H(f, uc)
    assert np.linalg.norm(uc-uh) < 1e-12
    T.destroy()

@pytest.mark.parametrize('axis', (0, 1, 2))
@pytest.mark.parametrize('family', ('chebyshev',))
def test_biharmonic3D(family, axis):
    la = cla
    N = (16, 16, 16)
    SD = FunctionSpace(N[allaxes3D[axis][0]], family=family, bc=(0, 0, 0, 0))
    K1 = FunctionSpace(N[allaxes3D[axis][1]], family='F', dtype='D')
    K2 = FunctionSpace(N[allaxes3D[axis][2]], family='F', dtype='d')
    subcomms = mpi4py_fft.pencil.Subcomm(MPI.COMM_WORLD, [0, 1, 1])
    bases = [0]*3
    bases[allaxes3D[axis][0]] = SD
    bases[allaxes3D[axis][1]] = K1
    bases[allaxes3D[axis][2]] = K2
    T = TensorProductSpace(subcomms, bases, axes=allaxes3D[axis])
    u = shenfun.TrialFunction(T)
    v = shenfun.TestFunction(T)
    mat = inner(v, div(grad(div(grad(u)))))

    H = la.Biharmonic(*mat)
    u = Function(T)
    u[:] = np.random.random(u.shape) + 1j*np.random.random(u.shape)
    f = Function(T)
    f = H.matvec(u, f)

    g0 = Function(T)
    g1 = Function(T)
    g2 = Function(T)
    mat = H.tpmats
    M = {d.get_key(): d for d in mat}
    g0 = M['SSBSBmat'].matvec(u, g0)
    g1 = M['ASBSBmat'].matvec(u, g1)
    g2 = M['BSBSBmat'].matvec(u, g2)
    T.destroy()

    assert np.linalg.norm(f-(g0+g1+g2)) < 1e-8, np.linalg.norm(f-(g0+g1+g2))

@pytest.mark.parametrize('axis', (0, 1))
@pytest.mark.parametrize('family', ('chebyshev',))
def test_biharmonic2D(family, axis):
    la = cla
    N = (16, 16)
    SD = FunctionSpace(N[axis], family=family, bc=(0, 0, 0, 0))
    K1 = FunctionSpace(N[(axis+1)%2], family='F', dtype='d')
    subcomms = mpi4py_fft.pencil.Subcomm(MPI.COMM_WORLD, allaxes2D[axis])
    bases = [K1]
    bases.insert(axis, SD)
    T = TensorProductSpace(subcomms, bases, axes=allaxes2D[axis])
    u = shenfun.TrialFunction(T)
    v = shenfun.TestFunction(T)
    mat = inner(v, div(grad(div(grad(u)))))

    H = la.Biharmonic(*mat)
    u = Function(T)
    u[:] = np.random.random(u.shape) + 1j*np.random.random(u.shape)
    f = Function(T)
    f = H.matvec(u, f)

    g0 = Function(T)
    g1 = Function(T)
    g2 = Function(T)
    mat = H.tpmats
    M = {d.get_key(): d for d in mat}
    g0 = M['SSBSBmat'].matvec(u, g0)
    g1 = M['ASBSBmat'].matvec(u, g1)
    g2 = M['BSBSBmat'].matvec(u, g2)
    T.destroy()
    assert np.linalg.norm(f-(g0+g1+g2)) < 1e-8


D = FunctionSpace(8, 'C', bc=(0, 0))
D0 = FunctionSpace(8, 'C')
F = FunctionSpace(8, 'F', dtype='d')
F2 = FunctionSpace(8, 'F', dtype='D')

@pytest.mark.parametrize('bases', ((D, D0), (D, F), (D, D0), (D, D0, F),
                                   (D, F2, F), (D0, F2, F)))
def test_blockmatrix(bases):
    T = TensorProductSpace(comm, bases)
    V = VectorSpace(T)
    u = shenfun.TrialFunction(V)
    v = shenfun.TestFunction(V)
    A = inner(u, v)
    B = BlockMatrix(A)
    uh = Function(V, val=1)
    c = Function(V)
    c = B.matvec(uh, c, use_scipy=True)
    c2 = Function(V)
    c2 = B.matvec(uh, c2, use_scipy=False)
    assert np.linalg.norm(c2-c) < 1e-8
    VQ = CompositeSpace([V, T])
    u, p = shenfun.TrialFunction(VQ)
    v, q = shenfun.TestFunction(VQ)
    A2 = inner(u, v) + [inner(p, q)]
    B2 = BlockMatrix(A2)
    uh = Function(VQ, val=1)
    c = Function(VQ)
    c = B2.matvec(uh, c, use_scipy=True)
    c2 = Function(VQ)
    c2 = B2.matvec(uh, c2, use_scipy=False)
    assert np.linalg.norm(c2-c) < 1e-8
    T.destroy()

@pytest.mark.parametrize('b0,b1', some_cbases2+some_lbases2)
def test_stencil(b0, b1):
    N = 14
    b0 = b0(N)
    b1 = b1(N)
    u = shenfun.TrialFunction(b1)
    v = shenfun.TestFunction(b0)
    B0 = inner(v, u, kind='vandermonde')
    B1 = inner(v, u, kind='stencil')
    C = B0-B1
    C.incorporate_scale()
    assert np.linalg.norm(C.diags('csr').data) < 1e-8
    B0 = inner(v*x, u, kind='vandermonde')
    B1 = inner(v*x, u, kind='stencil')
    C = B0-B1
    C.incorporate_scale()
    assert np.linalg.norm(C.diags('csr').data) < 1e-8

@pytest.mark.parametrize('b0,b1', some_lbases2+some_cbases2)
def test_adaptive_quadrature(b0, b1):
    N = 12
    b0 = b0(N)
    b1 = b1(N)
    u = shenfun.TrialFunction(b1)
    v = shenfun.TestFunction(b0)
    B0 = inner(v, u, assemble='adaptive')
    B1 = inner(v, u, assemble='quadrature', kind='stencil')
    C = B0-B1
    C.incorporate_scale()
    assert np.linalg.norm(C.diags('csr').data) < 1e-8

# Some bases take too long for Chebyshev, so just choose these three
some_cbases3 = [(cbases.ShenDirichlet, cbases.ShenDirichlet),
                (cbases.ShenDirichlet, cbases.ShenNeumann),
                (cbases.LowerDirichlet, cbases.UpperDirichlet)]
some_lbases3 = [lbases2[i] for i in np.random.randint(0, len(lbases2), 3)]

@pytest.mark.parametrize('b0,b1', some_lbases3+some_cbases3)
def test_exact(b0, b1):
    N = 10
    b0 = b0(N)
    b1 = b1(N)
    u = shenfun.TrialFunction(b1)
    v = shenfun.TestFunction(b0)
    B0 = inner(v, u, assemble='exact')
    B1 = inner(v, u, fixed_resolution=20)
    C = B0-B1
    C.incorporate_scale()
    assert np.linalg.norm(C.diags('csr').data) < 1e-8

if __name__ == '__main__':
    import sympy as sp
    x = sp.symbols('x', real=True)
    xp = sp.Symbol('x', real=True, positive=True)

    #test_exact(ltestBasis[-3], ltrialBasis[8])
    #test_mat(((ctestBasis[1], 0), (ctestBasis[1], 2), 1-x**2), cmatrices.ASDSDmatW, 'GC')
    #test_mat(((cbases.Orthogonal, 0), (lbases.Orthogonal, 0)), cmatrices.BTLmat, 'GL')
    #test_cmatvec(ctrialBasis[1], ctrialBasis[1], 'GC', 0)
    #test_lmatvec(ltrialBasis[0], ltrialBasis[0], 'LG', 2, 0)
    #test_lagmatvec(lagBasis[0], lagBasis[1], 'LG', 'python', 3, 2, 0)
    #test_hmatvec(hBasis[0], hBasis[0], 'HG', 'self', 3, 1, 1)
    #test_jmatvec(jBasis[5], jBasis[5], 'JG', 'self', 1, 1)
    #test_isub(((cbases.ShenNeumann, 0), (cbases.ShenDirichlet, 1)), cmatrices.CSNSDmat, 'GC')
    #test_add(((cbases.Orthogonal, 0), (cbases.ShenDirichlet, 1)), cmatrices.CTSDmat, 'GC')
    #test_blockmatrix((D, F2, F))
    #test_sub(*mats_and_quads[15])
    #test_mul2()
    #test_div2(cBasis[1], 'GC')
    #test_helmholtz2D('chebyshev', 1)
    #test_helmholtz3D('chebyshev', 0)
    #test_biharmonic3D('chebyshev', 0)
    #test_biharmonic2D('jacobi', 0)
    #test_stencil(cbases.Phi6, cbases.Compact3)
    test_adaptive_quadrature(lbases.Phi2, lbases.ShenNeumann)
